import os
import json

from datetime import datetime
from werkzeug.local import LocalProxy
from json.decoder import JSONDecodeError


class Gobits:
    """
    A small class that gathers information (bits) for a pub/sub message payload.

    Attributes:
        _request                Holds information about the cloud function request.
        _context                Holds information about the cloud function context.
        envelope                The envelope containing a pub/sub message.
        message                 The pub/sub message to be processed.
        processed               Time of processing (UTC).
        gcp_project             The source GCP project.
        execution_id            The id of the execution.
        execution_type          Type of the service processing the message.
        execution_trigger_type  Type of the trigger invoking the processor.
        function_name           Name of the processing cloud function.
        function_version        Version of the processing cloud function.
        event_id                The id of the trigger event.
        message_id              The pub/sub message id.
        message_publish_time    Time of publishing to pub/sub (UTC).
        source_subscription     The originating pub/sub subscription.
    """

    def __init__(self, request: LocalProxy = None, context=None):
        self._request = request
        self._context = context

    @property
    def request(self):
        return self._request

    @request.setter
    def request(self, value):
        self._request = value

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        self._context = value

    @property
    def envelope(self):
        if self._request:
            try:
                return json.loads(self._request.data.decode('utf-8'))
            except JSONDecodeError:
                return {}
        else:
            return {}

    @property
    def message(self):
        return self.envelope.get('message', {})

    @property
    def processed(self):
        return datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'

    @property
    def gcp_project(self):
        return os.getenv('X_GOOGLE_GCP_PROJECT')

    @property
    def execution_id(self):
        if self._request:
            return self._request.headers.get('Function-Execution-Id')
        return None

    @property
    def execution_type(self):
        if os.getenv('X_GOOGLE_FUNCTION_NAME'):
            return 'cloud_function'

        if os.getenv('BUILDER_OUTPUT'):
            return 'cloud_build'

        if os.getenv('GAE_APPLICATION'):
            return 'google_app_engine'

        return None

    @property
    def execution_trigger_type(self):
        return os.getenv('FUNCTION_TRIGGER_TYPE')

    @property
    def function_name(self):
        return os.getenv('X_GOOGLE_FUNCTION_NAME')

    @property
    def function_version(self):
        return os.getenv('X_GOOGLE_FUNCTION_VERSION')

    @property
    def event_id(self):
        if self._context:
            return self._context.event_id
        return None

    @property
    def message_id(self):
        return self.message.get('messageId')

    @property
    def message_publish_time(self):
        return self.message.get('publishTime')

    @property
    def source_subscription(self):
        return self.envelope.get('subscription')

    @classmethod
    def from_request(cls, request):
        return cls(request=request)

    @classmethod
    def from_context(cls, context):
        return cls(context=context)

    def to_json(self):

        gobits = {
            'processed': self.processed,
            'gcp_project': self.gcp_project,
            'execution_id': self.execution_id,
            'execution_type': self.execution_type,
            'execution_trigger_type': self.execution_trigger_type,
            'function_name': self.function_name,
            'function_version': self.function_version,
            'event_id': self.event_id,
            'message_id': self.message_id,
            'message_publish_time': self.message_publish_time,
            'source_subscription': self.source_subscription
        }

        response = {k: v for (k, v) in gobits.items() if v}

        return response



import os
import json
import unittest

from gobits import Gobits
from unittest.mock import Mock, patch  # noqa: F401
from werkzeug.local import LocalProxy  # noqa: F401

X_GOOGLE_GCP_PROJECT = 'my-test-project'
X_GOOGLE_FUNCTION_NAME = 'my-test-function'
SUBSCRIPTION = 'my-test-subscription'
FUNCTION_TRIGGER_TYPE = 'http'
X_GOOGLE_FUNCTION_VERSION = '99'
HTTP_FUNCTION_EXECUTION_ID = '0123456789'
EXECUTION_TYPE = 'cloud_function'
EVENT_ID = '1234567890123456'
MESSAGE_ID = '1234567890123456'
MESSAGE_PUBLISH_TIME = '2020-12-31T23:59:59.999Z'
BUILDER_OUTPUT = '/builder/outputs'


class TestGobits(unittest.TestCase):

    def setUp(self):
        os.environ['X_GOOGLE_GCP_PROJECT'] = X_GOOGLE_GCP_PROJECT
        os.environ['X_GOOGLE_FUNCTION_NAME'] = X_GOOGLE_FUNCTION_NAME
        os.environ['FUNCTION_TRIGGER_TYPE'] = FUNCTION_TRIGGER_TYPE
        os.environ['X_GOOGLE_FUNCTION_VERSION'] = X_GOOGLE_FUNCTION_VERSION
        self.gobits = Gobits()

    def test_processed(self):
        self.assertEqual(len(self.gobits.processed), 24),

    def test_gcp_project(self):
        self.assertEqual(self.gobits.gcp_project, X_GOOGLE_GCP_PROJECT)

    def test_execution_type(self):
        self.assertEqual(self.gobits.execution_type, EXECUTION_TYPE)

    def test_execution_trigger_type(self):
        self.assertEqual(self.gobits.execution_trigger_type, FUNCTION_TRIGGER_TYPE)

    def test_function_name(self):
        self.assertEqual(self.gobits.function_name, X_GOOGLE_FUNCTION_NAME)

    def test_function_version(self):
        self.assertEqual(self.gobits.function_version, X_GOOGLE_FUNCTION_VERSION)

    def test_to_json(self):
        gobits = self.gobits.to_json()
        self.assertEqual(gobits['gcp_project'], X_GOOGLE_GCP_PROJECT)
        self.assertEqual(gobits['execution_type'], EXECUTION_TYPE)
        self.assertEqual(gobits['execution_trigger_type'], FUNCTION_TRIGGER_TYPE)
        self.assertEqual(gobits['function_name'], X_GOOGLE_FUNCTION_NAME)
        self.assertEqual(gobits['function_version'], X_GOOGLE_FUNCTION_VERSION)

    def test_json_length(self):
        gobits = self.gobits.to_json()
        self.assertEqual(len(gobits), 6)

    def tearDown(self):
        os.environ['X_GOOGLE_GCP_PROJECT'] = ''
        os.environ['X_GOOGLE_FUNCTION_NAME'] = ''
        os.environ['FUNCTION_TRIGGER_TYPE'] = ''
        os.environ['X_GOOGLE_FUNCTION_VERSION'] = ''


class TestEmptyRequestGobits(unittest.TestCase):

    @patch('werkzeug.local.LocalProxy')
    def setUp(self, mock_request):
        mock_request.data = b''
        mock_request.headers = {'Function-Execution-Id': HTTP_FUNCTION_EXECUTION_ID}
        self.gobits = Gobits.from_request(request=mock_request)

    def test_to_json(self):
        gobits = self.gobits.to_json()
        self.assertEqual(gobits['execution_id'], HTTP_FUNCTION_EXECUTION_ID)


class TestRequestGobits(unittest.TestCase):

    @patch('werkzeug.local.LocalProxy')
    def setUp(self, mock_request):
        mock_envelope = dict(
            subscription=SUBSCRIPTION,
            message=dict(
                messageId=MESSAGE_ID,
                publishTime=MESSAGE_PUBLISH_TIME
            )
        )
        mock_request.data = json.dumps(mock_envelope).encode('utf-8')
        mock_request.headers = {'Function-Execution-Id': HTTP_FUNCTION_EXECUTION_ID}
        self.gobits = Gobits.from_request(request=mock_request)

    def test_processed(self):
        self.assertEqual(len(self.gobits.processed), 24),

    def test_execution_id(self):
        self.assertEqual(self.gobits.execution_id, HTTP_FUNCTION_EXECUTION_ID)

    def test_message_id(self):
        self.assertEqual(self.gobits.message_id, MESSAGE_ID)

    def test_message_publish_time(self):
        self.assertEqual(self.gobits.message_publish_time, MESSAGE_PUBLISH_TIME)

    def test_to_json(self):
        gobits = self.gobits.to_json()
        self.assertEqual(gobits['execution_id'], HTTP_FUNCTION_EXECUTION_ID)
        self.assertEqual(gobits['message_id'], MESSAGE_ID)
        self.assertEqual(gobits['message_publish_time'], MESSAGE_PUBLISH_TIME)

    def test_json_length(self):
        gobits = self.gobits.to_json()
        self.assertEqual(len(gobits), 5)


class TestContextGobits(unittest.TestCase):

    @patch('unittest.mock.Mock')
    def setUp(self, mock_context):
        mock_context.event_id = EVENT_ID
        self.gobits = Gobits.from_context(context=mock_context)

    def test_processed(self):
        self.assertEqual(len(self.gobits.processed), 24),

    def test_event_id(self):
        self.assertEqual(self.gobits.event_id, EVENT_ID)

    def test_to_json(self):
        gobits = self.gobits.to_json()
        self.assertEqual(gobits['event_id'], EVENT_ID)

    def test_json_length(self):
        gobits = self.gobits.to_json()
        self.assertEqual(len(gobits), 2)


class TestCloudBuildGobits(unittest.TestCase):

    def setUp(self):
        os.environ['BUILDER_OUTPUT'] = BUILDER_OUTPUT
        self.gobits = Gobits()

    def test_processed(self):
        self.assertEqual(len(self.gobits.processed), 24),

    def test_to_json(self):
        gobits = self.gobits.to_json()
        self.assertEqual(gobits['execution_type'], 'cloud_build')

    def test_json_length(self):
        gobits = self.gobits.to_json()
        self.assertEqual(len(gobits), 2)

    def tearDown(self):
        os.environ['BUILDER_OUTPUT'] = ''


if __name__ == '__main__':
    unittest.main()



import os

from setuptools import setup

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

with open('test-requirements.txt') as f:
    test_requires = f.read().splitlines()

setup(
    name='gobits',
    version=os.getenv('TAG_NAME', '0.0.0'),
    description='Small Python package to add GCP metadata to pub/sub messages',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/vwt-digital/gobits',
    author='VWT Digital',
    author_email='support@vwt.digital',
    license='GPLv3+',
    packages=['gobits'],
    install_requires=install_requires,
    test_suite="tests",
    tests_require=test_requires,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering'
    ]
)
